﻿using System;
using System.Collections.Generic;
using System.Linq;

namespace KnapsackProblem
{
    public class Program
    {
        public static void Main(string[] args)
        {
            List<Item> items = new List<Item>();

            items.Add(new Item(7, 42));
            items.Add(new Item(3, 12));
            items.Add(new Item(4, 40));
            items.Add(new Item(5, 25));

            Console.WriteLine("List of avialable Items:");
            Console.WriteLine(string.Join("\n", items));

            int knapsackCapacity = 10;
            Console.WriteLine($"Knapsack capacity: {knapsackCapacity}");


            Console.WriteLine("\nGreedy Solution:");
            KnapsackGreedy(items, knapsackCapacity);
            Console.WriteLine("\nDynamic Solution:");
            KnapsackDynamic(items, knapsackCapacity);

            Console.ReadLine();
        }

        public static void KnapsackGreedy(List<Item> items, int knapsackCapacity)
        {
            List<Item> knapsack = new List<Item>();

            List<Item> avialable = new List<Item>();
            avialable.AddRange(items);
            avialable = avialable.OrderByDescending(item => item.ValuePerWeight())
                                 .ThenBy(item => item.Weight)
                                 .ToList();

            while (avialable.Count > 0)
            {
                // Items are sorted by highest value then lowest weight.
                // This ensures that the first hit is potentiall the most profitable.
                Item selected = avialable.First();
                avialable.Remove(selected);

                knapsack.Add(selected);
                knapsackCapacity -= selected.Weight;

                // Keep only the items whose weights are
                // less than or equal to our remaining capacity.
                avialable = avialable.Where(item => item.Weight <= knapsackCapacity).ToList();
            }

            Console.WriteLine(string.Join("\n", knapsack));
            Console.WriteLine($"Max Value: {knapsack.Select(item => item.Value).Sum()}");
        }

        public static void KnapsackDynamic(List<Item> items, int knapsackCapacity)
        {
            // Assume every item has a weight equal or less than the capacity.
            List<Item> sortedItems = items.OrderBy(item => item.Weight).ToList();
            int itemCount = sortedItems.Count;
            int[,] profitMatrix = new int[itemCount + 1, knapsackCapacity + 1];

            // Each row is an item.
            // Each column is the max weight.
            // Item 0 is a placeholder, cannot be picked.
            // Column 0 means max weight 0, can't store anything.
            // So we skip the first row and column, leaving them to be 0s.
            for (int row = 1; row <= itemCount; row++)
            {
                int realIndex = row - 1;
                Item currentItem = sortedItems[realIndex];

                for (int column = 1; column <= knapsackCapacity; column++)
                {
                    // Check if current item fits in this column
                    if (currentItem.Weight <= column)
                    {
                        // See how much capacity is left over
                        int leftOverCapacity = column - currentItem.Weight;

                        // Get the max value for the left over capacity
                        // Check only the previous row because we can't
                        // take the same item twice.
                        int leftOverCapacityMaxValue = profitMatrix[row - 1, leftOverCapacity];

                        // Add current item's value and whatever we can
                        // get for the left over capacity to get max
                        // potential for the current max weight for all
                        // items up until the current one.
                        int A = currentItem.Value + leftOverCapacityMaxValue;

                        // Get the last set of combination of items.
                        int B = profitMatrix[row - 1, column];

                        // If the previous row had a better value, just copy that.
                        // If not, current combo is kept.
                        profitMatrix[row, column] = Math.Max(A, B);
                    }
                    else
                    {
                        // Not enough capacity to fit new item.
                        // So we copy from above.
                        profitMatrix[row, column] = profitMatrix[row - 1, column];
                    }
                }
            }

            // Now we have a matrix with the best items to pick for
            // every value of max weight, the top value score is in
            // the bottom right of the matrix.
            Console.WriteLine($"Max value: {profitMatrix[itemCount, knapsackCapacity]}");

            // We can find out which items we picked by looking them up
            // bottom to top, back to front:

            int searchColumn = knapsackCapacity;
            int searchRow = itemCount;
            while (true)
            {
                int currentValue = profitMatrix[searchRow, searchColumn];
                int previousValue = profitMatrix[searchRow -1, searchColumn];

                // If the current value 0, we are likely done searching.
                if (currentValue == 0)
                {
                    break;
                }
                else if (currentValue == previousValue)
                {
                    // If current value and the one above are the same
                    // It's either copied or they are the same.
                    // Either way, we assume that the current item is
                    // not optimal.
                    searchRow--;
                }
                else
                {
                    // We found a participating item, print it and find out the next one.
                    // Remember the offset.
                    Console.WriteLine(sortedItems[searchRow - 1]);
                    int itemWeight = sortedItems[searchRow - 1].Weight;
                    // subtrach the weight of the participating item,
                    // from the left over, and go up a row.
                    searchColumn = searchColumn - itemWeight;
                    searchRow--;
                }
            }

        }
    }

    public class Item
    {
        public int Weight { get; private set; }
        public int Value { get; private set; }

        public double ValuePerWeight()
        {
            return (double)Value / (double)Weight;
        }

        public Item(int w, int v)
        {
            Weight = w;
            Value = v;
        }

        public override string ToString()
        {
            return $"W:{Weight} V:{Value} V/W:{ValuePerWeight()}";
        }
    }
}